@@ -1,7 +1,5 @@
 # encoding: utf-8
 
-require "amq/client/exchange"
-
 module AMQP
 
   # h2. What are AMQP exchanges?
@@ -50,7 +48,7 @@ module AMQP
   # As part of the standard, the server _must_ predeclare the direct exchange
   # 'amq.direct' and the fanout exchange 'amq.fanout' (all exchange names
   # starting with 'amq.' are reserved). Attempts to declare an exchange using
-  # 'amq.' as the name will raise an AMQP::Error and fail. In practice these
+  # 'amq.' as the name will result in a channel-level exception and fail. In practice these
   # default exchanges are never used directly by client code.
   #
   #
@@ -134,16 +132,26 @@ module AMQP
   # @see Channel#topic
   # @see Channel#headers
   # @see Queue
-  # @see http://bit.ly/hw2ELX AMQP 0.9.1 specification (Section 2.1.1)
-  # @see http://bit.ly/hw2ELX AMQP 0.9.1 specification (Section 2.1.5)
-  # @see http://bit.ly/hw2ELX AMQP 0.9.1 specification (Section 3.1.3)
-  class Exchange < AMQ::Client::Exchange
+  # @see http://files.travis-ci.org/docs/amqp/0.9.1/AMQP091Specification.pdf AMQP 0.9.1 specification (Section 2.1.1)
+  # @see http://files.travis-ci.org/docs/amqp/0.9.1/AMQP091Specification.pdf AMQP 0.9.1 specification (Section 2.1.5)
+  # @see http://files.travis-ci.org/docs/amqp/0.9.1/AMQP091Specification.pdf AMQP 0.9.1 specification (Section 3.1.3)
+  class Exchange
+
+
+    #
+    # Behaviours
+    #
+
+    include Entity
+    extend  ProtocolMethodHandlers
+
 
     #
     # API
     #
 
     DEFAULT_CONTENT_TYPE = "application/octet-stream".freeze
+    BUILTIN_TYPES = [:fanout, :direct, :topic, :headers].freeze
 
 
     # The default exchange. Default exchange is a direct exchange that is predefined.
@@ -162,7 +170,7 @@ module AMQP
     #   AMQP::Exchange.default(channel).publish("make clean", routing_key => "tasks")
     #
     # @see Exchange
-    # @see http://bit.ly/hw2ELX AMQP 0.9.1 specification (Section 2.1.2.4)
+    # @see http://files.travis-ci.org/docs/amqp/0.9.1/AMQP091Specification.pdf AMQP 0.9.1 specification (Section 2.1.2.4)
     # @note Do not confuse default exchange with amq.direct: amq.direct is a pre-defined direct
     #       exchange that doesn't have any special routing semantics.
     # @return [Exchange] An instance that corresponds to the default exchange (of type direct).
@@ -191,6 +199,13 @@ module AMQP
     #                 from the broker arrives.
     attr_accessor :on_declare
 
+    # Channel this exchange belongs to.
+    attr_reader :channel
+
+    # @return [Hash] Additional arguments given on queue declaration. Typically used by AMQP extensions.
+    attr_reader :arguments
+
+
     # @return [String]
     attr_reader :default_routing_key
     alias key default_routing_key
@@ -278,10 +293,6 @@ module AMQP
     # @option opts [Hash] :arguments (nil)  A hash of optional arguments with the declaration. Some brokers implement
     #                                          AMQP extensions using x-prefixed declaration arguments.
     #
-    #
-    # @raise [AMQP::Error] Raised when exchange is redeclared with parameters different from original declaration.
-    # @raise [AMQP::Error] Raised when exchange is declared with :passive => true and the exchange does not exist.
-    #
     # @yield [exchange, declare_ok] Yields successfully declared exchange instance and AMQP method (exchange.declare-ok) instance. The latter is optional.
     # @yieldparam [Exchange] exchange Exchange that is successfully declared and is ready to be used.
     # @yieldparam [AMQP::Protocol::Exchange::DeclareOk] declare_ok AMQP exchange.declare-ok) instance.
@@ -292,7 +303,7 @@ module AMQP
     # @see Channel#topic
     # @see Channel#headers
     # @see Queue
-    # @see http://bit.ly/hw2ELX AMQP 0.9.1 specification (Section 3.1.3)
+    # @see http://files.travis-ci.org/docs/amqp/0.9.1/AMQP091Specification.pdf AMQP 0.9.1 specification (Section 3.1.3)
     #
     # @return [Exchange]
     # @api public
@@ -305,18 +316,30 @@ module AMQP
 
       @status                  = :unknown
       @default_publish_options = (opts.delete(:default_publish_options) || {
-                                    :routing_key  => @default_routing_key,
-                                    :mandatory    => false,
-                                    :immediate    => false
-                                  }).freeze
+          :routing_key  => @default_routing_key,
+          :mandatory    => false
+        }).freeze
 
       @default_headers = (opts.delete(:default_headers) || {
-                            :content_type => DEFAULT_CONTENT_TYPE,
-                            :persistent   => false,
-                            :priority     => 0
-                          }).freeze
+          :content_type => DEFAULT_CONTENT_TYPE,
+          :persistent   => false,
+          :priority     => 0
+        }).freeze
 
-      super(channel.connection, channel, name, type)
+      if !(BUILTIN_TYPES.include?(type.to_sym) || type.to_s =~ /^x-.+/i)
+        raise UnknownExchangeTypeError.new(BUILTIN_TYPES, type)
+      end
+
+      @channel    = channel
+      @name       = name
+      @type       = type
+
+      # register pre-declared exchanges
+      if @name == AMQ::Protocol::EMPTY_STRING || @name =~ /^amq\.(direct|fanout|topic|match|headers)/
+        @channel.register_exchange(self)
+      end
+
+      super(channel.connection)
 
       # The AMQP 0.8 specification (as well as 0.9.1) in 1.1.4.2 mentiones
       # that Exchange.Declare-Ok confirms the name of the exchange (because
@@ -342,9 +365,9 @@ module AMQP
                 end
               end
 
-              self.declare(passive = @opts[:passive], durable = @opts[:durable], auto_delete = @opts[:auto_delete], nowait = @opts[:nowait], @opts[:arguments], &shim)
+              self.exchange_declare(@opts[:passive], @opts[:durable], @opts[:auto_delete], @opts[:internal], @opts[:nowait], @opts[:arguments], &shim)
             else
-              self.declare(passive = @opts[:passive], durable = @opts[:durable], auto_delete = @opts[:auto_delete], nowait = @opts[:nowait], @opts[:arguments])
+              self.exchange_declare(@opts[:passive], @opts[:durable], @opts[:auto_delete], @opts[:internal], @opts[:nowait], @opts[:arguments])
             end
           end
         end
@@ -364,6 +387,46 @@ module AMQP
       @channel
     end
 
+    # @return [Boolean] true if this exchange is of type `fanout`
+    # @api public
+    def fanout?
+      @type == :fanout
+    end
+
+    # @return [Boolean] true if this exchange is of type `direct`
+    # @api public
+    def direct?
+      @type == :direct
+    end
+
+    # @return [Boolean] true if this exchange is of type `topic`
+    # @api public
+    def topic?
+      @type == :topic
+    end
+
+    # @return [Boolean] true if this exchange is of type `headers`
+    # @api public
+    def headers?
+      @type == :headers
+    end
+
+    # @return [Boolean] true if this exchange is of a custom type (begins with x-)
+    # @api public
+    def custom_type?
+      @type.to_s =~ /^x-.+/i
+    end # custom_type?
+
+    # @return [Boolean] true if this exchange is a pre-defined one (amq.direct, amq.fanout, amq.match and so on)
+    def predefined?
+      @name && ((@name == AMQ::Protocol::EMPTY_STRING) || !!(@name =~ /^amq\.(direct|fanout|topic|headers|match)/i))
+    end # predefined?
+
+    # @return [Boolean] true if this exchange is an internal exchange
+    def internal?
+      @opts[:internal]
+    end
+
     # Publishes message to the exchange. The message will be routed to queues by the exchange
     # and distributed to any active consumers. Routing logic is determined by exchange type and
     # configuration as well as message attributes (like :routing_key or message headers).
@@ -443,12 +506,6 @@ module AMQP
     #                                      unroutable message back to the client with basic.return AMQPmethod.
     #                                      If message is not mandatory, the server silently drops the message.
     #
-    # @option options [Boolean] :immediate (false) This flag tells the server how to react if the message cannot be
-    #                                      routed to a queue consumer immediately.  If this flag is set, the
-    #                                      server will return an undeliverable message with a Return method.
-    #                                      If this flag is zero, the server will queue the message, but with
-    #                                      no guarantee that it will ever be consumed.
-    #
     # @option options [Boolean] :persistent (false) When true, this message will be persisted to disk and remain in the queue until
     #                                       it is consumed. When false, the message is only kept in a transient store
     #                                       and will lost in case of server restart.
@@ -479,9 +536,9 @@ module AMQP
       @channel.once_open do
         properties                 = @default_headers.merge(options)
         properties[:delivery_mode] = properties.delete(:persistent) ? 2 : 1
-        super(payload.to_s, opts[:key] || opts[:routing_key] || @default_routing_key, properties, opts[:mandatory], opts[:immediate])
+        basic_publish(payload.to_s, opts[:key] || opts[:routing_key] || @default_routing_key, properties, opts[:mandatory])
 
-        # don't pass block to AMQ::Client::Exchange#publish because it will be executed
+        # don't pass block to AMQP::Exchange#publish because it will be executed
         # immediately and we want to do it later. See ruby-amqp/amqp/#67 MK.
         EventMachine.next_tick(&block) if block
       end
@@ -510,7 +567,7 @@ module AMQP
     # @api public
     def delete(opts = {}, &block)
       @channel.once_open do
-        super(opts.fetch(:if_unused, false), opts.fetch(:nowait, false), &block)
+        exchange_delete(opts.fetch(:if_unused, false), opts.fetch(:nowait, false), &block)
       end
 
       # backwards compatibility
@@ -548,11 +605,323 @@ module AMQP
     end
 
 
+    # @group Declaration
+
+    # @api public
+    def exchange_declare(passive = false, durable = false, auto_delete = false, internal = false, nowait = false, arguments = nil, &block)
+      # for re-declaration
+      @passive     = passive
+      @durable     = durable
+      @auto_delete = auto_delete
+      @arguments   = arguments
+      @internal    = internal
+
+      @connection.send_frame(AMQ::Protocol::Exchange::Declare.encode(@channel.id, @name, @type.to_s, passive, durable, auto_delete, internal, nowait, arguments))
+
+      unless nowait
+        self.define_callback(:declare, &block)
+        @channel.exchanges_awaiting_declare_ok.push(self)
+      end
+
+      self
+    end
+
+
+    # @api public
+    def redeclare(&block)
+      nowait = block.nil?
+      @connection.send_frame(AMQ::Protocol::Exchange::Declare.encode(@channel.id, @name, @type.to_s, @passive, @durable, @auto_delete, @internal, nowait, @arguments))
+
+      unless nowait
+        self.define_callback(:declare, &block)
+        @channel.exchanges_awaiting_declare_ok.push(self)
+      end
+
+      self
+    end # redeclare(&block)
+
+    # @endgroup
+
+
+    # @api public
+    def exchange_delete(if_unused = false, nowait = false, &block)
+      @connection.send_frame(AMQ::Protocol::Exchange::Delete.encode(@channel.id, @name, if_unused, nowait))
+
+      unless nowait
+        self.define_callback(:delete, &block)
+
+        # TODO: delete itself from exchanges cache
+        @channel.exchanges_awaiting_delete_ok.push(self)
+      end
+
+      self
+    end # delete(if_unused = false, nowait = false)
+
+
+    # @group Exchange to Exchange Bindings
+
+    # This method binds a source exchange to a destination exchange. Messages
+    # sent to the source exchange are forwarded to the destination exchange, provided
+    # the message routing key matches the routing key specified when binding the
+    # exchanges.
+    #
+    # A valid exchange name (or reference) must be passed as the first
+    # parameter.
+    # @example Binding two source exchanges to a destination exchange
+    #
+    #  ch          = AMQP::Channel.new(connection)
+    #  source1     = ch.fanout('backlog-events-datacenter-1')
+    #  source2     = ch.fanout('backlog-events-datacenter-2')
+    #  destination = ch.fanout('baklog-events')
+    #  destination.bind(source1).bind(source2)
+    #
+    #
+    # @param [Exchange] Exchange to bind to. May also be a string or any object that responds to #name.
+    #
+    # @option opts [String] :routing_key   Specifies the routing key for the binding. The routing key is
+    #                                      used for routing messages depending on the exchange configuration.
+    #                                      Not all exchanges use a routing key! Refer to the specific
+    #                                      exchange documentation.
+    #
+    # @option opts [Hash] :arguments (nil)  A hash of optional arguments with the declaration. Headers exchange type uses these metadata
+    #                                       attributes for routing matching.
+    #                                       In addition, brokers may implement AMQP extensions using x-prefixed declaration arguments.
+    #
+    # @option opts [Boolean] :nowait (true)  If set, the server will not respond to the method. The client should
+    #                                       not wait for a reply method.  If the server could not complete the
+    #                                       method it will raise a channel or connection exception.
+    # @return [Exchange] Self
+    #
+    #
+    # @yield [] Since exchange.bind-ok carries no attributes, no parameters are yielded to the block.
+    #
+
+    # @api public
+    def bind(source, opts = {}, &block)
+      source = source.name if source.respond_to?(:name)
+      routing_key = opts[:key] || opts[:routing_key] || AMQ::Protocol::EMPTY_STRING
+      arguments = opts[:arguments] || {}
+      nowait = opts[:nowait] || block.nil?
+      @channel.once_open do
+        @connection.send_frame(AMQ::Protocol::Exchange::Bind.encode(@channel.id, @name, source, routing_key, nowait, arguments))
+        unless nowait
+          self.define_callback(:bind, &block)
+          @channel.exchanges_awaiting_bind_ok.push(self)
+        end
+      end
+      self
+    end
+
+    # This method unbinds a source exchange from a previously bound destination exchange.
+    #
+    # A valid exchange name (or reference) must be passed as the first
+    # parameter.
+    # @example Binding and unbinding two exchanges
+    #
+    #  ch          = AMQP::Channel.new(connection)
+    #  source      = ch.fanout('backlog-events')
+    #  destination = ch.fanout('backlog-events-copy')
+    #  destination.bind(source)
+    #  ...
+    #  destination.unbind(source)
+    #
+    #
+    # @param [Exchange] Exchange to bind to. May also be a string or any object that responds to #name.
+    #
+    # @option opts [String] :routing_key   Specifies the routing key for the binding. The routing key is
+    #                                      used for routing messages depending on the exchange configuration.
+    #                                      Not all exchanges use a routing key! Refer to the specific
+    #                                      exchange documentation.
+    #
+    # @option opts [Hash] :arguments (nil)  A hash of optional arguments with the declaration. Headers exchange type uses these metadata
+    #                                       attributes for routing matching.
+    #                                       In addition, brokers may implement AMQP extensions using x-prefixed declaration arguments.
+    #
+    # @option opts [Boolean] :nowait (true)  If set, the server will not respond to the method. The client should
+    #                                       not wait for a reply method.  If the server could not complete the
+    #                                       method it will raise a channel or connection exception.
+    # @return [Exchange] Self
+    #
+    #
+    # @yield [] Since exchange.unbind-ok carries no attributes, no parameters are yielded to the block.
+    #
+
+    # @api public
+    def unbind(source, opts = {}, &block)
+      source = source.name if source.respond_to?(:name)
+      routing_key = opts[:key] || opts[:routing_key] || AMQ::Protocol::EMPTY_STRING
+      arguments = opts[:arguments] || {}
+      nowait = opts[:nowait] || block.nil?
+      @channel.once_open do
+        @connection.send_frame(AMQ::Protocol::Exchange::Unbind.encode(@channel.id, @name, source, routing_key, nowait, arguments))
+        unless nowait
+          self.define_callback(:unbind, &block)
+          @channel.exchanges_awaiting_unbind_ok.push(self)
+        end
+      end
+      self
+    end
+
+    # @endgroup
+
+    # @group Publishing Messages
+
+    # @api public
+    def basic_publish(payload, routing_key = AMQ::Protocol::EMPTY_STRING, user_headers = {}, mandatory = false)
+      headers = { :priority => 0, :delivery_mode => 2, :content_type => "application/octet-stream" }.merge(user_headers)
+      @connection.send_frameset(AMQ::Protocol::Basic::Publish.encode(@channel.id, payload, headers, @name, routing_key, mandatory, false, @connection.frame_max), @channel)
+
+      # publisher confirms support. MK.
+      @channel.exec_callback(:after_publish)
+      self
+    end
+
+
+    # @api public
+    def on_return(&block)
+      self.redefine_callback(:return, &block)
+
+      self
+    end # on_return(&block)
+
+    # @endgroup
+
+
+
+    # @group Error Handling and Recovery
+
+
+    # Defines a callback that will be executed after TCP connection is interrupted (typically because of a network failure).
+    # Only one callback can be defined (the one defined last replaces previously added ones).
+    #
+    # @api public
+    def on_connection_interruption(&block)
+      self.redefine_callback(:after_connection_interruption, &block)
+    end # on_connection_interruption(&block)
+    alias after_connection_interruption on_connection_interruption
+
+    # @private
+    def handle_connection_interruption(method = nil)
+      self.exec_callback_yielding_self(:after_connection_interruption)
+    end # handle_connection_interruption
+
+
+
+    # Defines a callback that will be executed after TCP connection is recovered after a network failure
+    # but before AMQP connection is re-opened.
+    # Only one callback can be defined (the one defined last replaces previously added ones).
+    #
+    # @api public
+    def before_recovery(&block)
+      self.redefine_callback(:before_recovery, &block)
+    end # before_recovery(&block)
+
+    # @private
+    def run_before_recovery_callbacks
+      self.exec_callback_yielding_self(:before_recovery)
+    end
+
+
+    # Defines a callback that will be executed when AMQP connection is recovered after a network failure..
+    # Only one callback can be defined (the one defined last replaces previously added ones).
+    #
+    # @api public
+    def on_recovery(&block)
+      self.redefine_callback(:after_recovery, &block)
+    end # on_recovery(&block)
+    alias after_recovery on_recovery
+
+    # @private
+    def run_after_recovery_callbacks
+      self.exec_callback_yielding_self(:after_recovery)
+    end
+
+
+    # Called by associated connection object when AMQP connection has been re-established
+    # (for example, after a network failure).
+    #
+    # @api plugin
+    def auto_recover
+      self.redeclare unless predefined?
+    end # auto_recover
+
+    # @endgroup
+
+
+
+    #
+    # Implementation
+    #
+
+
+    def handle_declare_ok(method)
+      @channel.register_exchange(self)
+
+      self.exec_callback_once_yielding_self(:declare, method)
+    end
+
+    def handle_bind_ok(method)
+      self.exec_callback_once(:bind, method)
+    end
+
+    def handle_unbind_ok(method)
+      self.exec_callback_once(:unbind, method)
+    end
+
+    def handle_delete_ok(method)
+      self.exec_callback_once(:delete, method)
+    end
+
+
+    self.handle(AMQ::Protocol::Exchange::DeclareOk) do |connection, frame|
+      method   = frame.decode_payload
+      channel  = connection.channels[frame.channel]
+      exchange = channel.exchanges_awaiting_declare_ok.shift
+
+      exchange.handle_declare_ok(method)
+    end # handle
+
+    self.handle(AMQ::Protocol::Exchange::BindOk) do |connection, frame|
+      method   = frame.decode_payload
+      channel  = connection.channels[frame.channel]
+      exchange = channel.exchanges_awaiting_bind_ok.shift
+
+      exchange.handle_bind_ok(method)
+    end # handle
+
+    self.handle(AMQ::Protocol::Exchange::UnbindOk) do |connection, frame|
+      method   = frame.decode_payload
+      channel  = connection.channels[frame.channel]
+      exchange = channel.exchanges_awaiting_unbind_ok.shift
+
+      exchange.handle_unbind_ok(method)
+    end # handle
+
+    self.handle(AMQ::Protocol::Exchange::DeleteOk) do |connection, frame|
+      channel  = connection.channels[frame.channel]
+      exchange = channel.exchanges_awaiting_delete_ok.shift
+      exchange.handle_delete_ok(frame.decode_payload)
+    end # handle
+
+
+    self.handle(AMQ::Protocol::Basic::Return) do |connection, frame, content_frames|
+      channel  = connection.channels[frame.channel]
+      method   = frame.decode_payload
+      exchange = channel.find_exchange(method.exchange)
+
+      header   = content_frames.shift
+      body     = content_frames.map { |frame| frame.payload }.join
+
+      exchange.exec_callback(:return, method, header, body)
+    end
+
+
+
     protected
 
     # @private
     def self.add_default_options(type, name, opts, block)
-      { :exchange => name, :type => type, :nowait => block.nil? }.merge(opts)
+      { :exchange => name, :type => type, :nowait => block.nil?, :internal => false }.merge(opts)
     end
   end # Exchange
 end # AMQP
